@@ -36,7 +36,7 @@ repository.
 
 =item B<-R>, B<--release-use-changelog>
 
-When used in conjunction with B<--release>, if there are uncommited
+When used in conjunction with B<--release>, if there are uncommitted
 changes to the changelog then derive the commit message from those
 changes rather than using the default message.
 
@@ -82,6 +82,11 @@ the message.
 This option is set by default and ignored if more than one line of
 the message begins with "[*+-] ".
 
+=item B<--sign-commit>, B<--no-sign-commit>
+
+If this option is set, then the commits that debcommit creates will be
+signed using gnupg. Currently this is only supported by git, hg, and bzr.
+
 =item B<--sign-tags>, B<--no-sign-tags>
 
 If this option is set, then tags that debcommit creates will be signed
@@ -116,6 +121,11 @@ command line parameter being used. The default is I<yes>.
 If this is set to I<yes>, then it is the same as the B<--sign-tags> command
 line parameter being used. The default is I<no>.
 
+=item B<DEBCOMMIT_SIGN_COMMITS>
+
+If this is set to I<yes>, then it is the same as the B<--sign-commit>
+command line parameter being used. The default is I<no>.
+
 =item B<DEBCOMMIT_RELEASE_USE_CHANGELOG>
 
 If this is set to I<yes>, then it is the same as the B<--release-use-changelog>
@@ -172,7 +182,7 @@ options to "bzr commit" will be generated to associate the revision and the bugs
 
 use warnings;
 use strict;
-use Getopt::Long qw(:config gnu_getopt);
+use Getopt::Long qw(:config bundling permute no_getopt_compat);
 use Cwd;
 use File::Basename;
 use File::Temp;
@@ -202,8 +212,10 @@ Options:
    -C --confirm        Ask for confirmation of the message before commit
    -e --edit           Edit the message in EDITOR before commit
    -a --all            Commit all files (default except for git)
-   -s --strip-message  Strip the leading '* ' from the commit message
-   --no-strip-message  Do not strip a leading '* ' (default)
+   -s --strip-message  Strip the leading '* ' from the commit message (default)
+   --no-strip-message  Do not strip a leading '* '
+   --sign-commit       Enable signing of the commit (git, hg, and bzr)
+   --no-sign-commit    Do not sign the commit (default)
    --sign-tags         Enable signing of tags (git only)
    --no-sign-tags      Do not sign tags (default)
    --changelog-info    Use author and date information from the changelog
@@ -240,6 +252,7 @@ my $confirm=0;
 my $edit=0;
 my $all=0;
 my $stripmessage=1;
+my $signcommit=0;
 my $signtags=0;
 my $changelog;
 my $changelog_info=0;
@@ -257,6 +270,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     my @config_files = ('/etc/devscripts.conf', '~/.devscripts');
     my %config_vars = (
 		       'DEBCOMMIT_STRIP_MESSAGE' => 'yes',
+		       'DEBCOMMIT_SIGN_COMMITS' => 'no',
 		       'DEBCOMMIT_SIGN_TAGS' => 'no',
 		       'DEBCOMMIT_RELEASE_USE_CHANGELOG' => 'no',
 		       'DEBSIGN_KEYID' => '',
@@ -278,6 +292,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     # Check validity
     $config_vars{'DEBCOMMIT_STRIP_MESSAGE'} =~ /^(yes|no)$/
 	or $config_vars{'DEBCOMMIT_STRIP_MESSAGE'}='yes';
+    $config_vars{'DEBCOMMIT_SIGN_COMMITS'} =~ /^(yes|no)$/
+	or $config_vars{'DEBCOMMIT_SIGN_COMMITS'}='no';
     $config_vars{'DEBCOMMIT_SIGN_TAGS'} =~ /^(yes|no)$/
 	or $config_vars{'DEBCOMMIT_SIGN_TAGS'}='no';
     $config_vars{'DEBCOMMIT_RELEASE_USE_CHANGELOG'} =~ /^(yes|no)$/
@@ -292,6 +308,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     chomp $modified_conf_msg;
 
     $stripmessage = $config_vars{'DEBCOMMIT_STRIP_MESSAGE'} eq 'no' ? 0 : 1;
+    $signcommit = $config_vars{'DEBCOMMIT_SIGN_COMMITS'} eq 'no' ? 0 : 1;
     $signtags = $config_vars{'DEBCOMMIT_SIGN_TAGS'} eq 'no' ? 0 : 1;
     $release_use_changelog = $config_vars{'DEBCOMMIT_RELEASE_USE_CHANGELOG'} eq 'no' ? 0 : 1;
     if (exists $config_vars{'DEBSIGN_KEYID'} &&
@@ -321,6 +338,7 @@ if (! GetOptions(
 		 "a|all" => \$all,
 		 "c|changelog=s" => \$changelog,
 		 "s|strip-message!" => \$stripmessage,
+		 "sign-commit!" => \$signcommit,
 		 "sign-tags!" => \$signtags,
 		 "changelog-info!" => \$changelog_info,
 		 "R|release-use-changelog!" => \$release_use_changelog,
@@ -430,7 +448,8 @@ sub getprog {
     if (-d ".bzr") {
 	return "bzr";
     }
-    if (-d ".git") {
+    if (-e ".git") {
+	# With certain forms of git checkouts, .git can be a file instead of a directory
 	return "git";
     }
     if (-d ".hg") {
@@ -456,7 +475,7 @@ sub getprog {
 	if (-d "$dir/.bzr") {
 	    return "bzr";
 	}
-	if (-d "$dir/.git") {
+	if (-e "$dir/.git") {
 	    return "git";
 	}
 	if (-d "$dir/.hg") {
@@ -514,6 +533,14 @@ sub commit {
 	$action_rc = $diffmode
 	    ? action($prog, "diff", @files_to_commit)
 	    : action($prog, "commit", "-m", $message, @extra_args, @files_to_commit);
+	if ($prog eq 'hg' && $action_rc && $signcommit) {
+	   my @sign_args;
+	   push(@sign_args, '-k', $keyid) if $keyid;
+	   push(@sign_args, '-u', $maintainer, '-d', $date) if $changelog_info;
+	   if (!action($prog, 'sign', @sign_args)) {
+	      die "$progname: failed to sign commit\n";
+	   }
+	}
     }
     elsif ($prog eq 'git') {
 	if (! @files_to_commit && ($all || $release)) {
@@ -535,6 +562,11 @@ sub commit {
 	    if ($changelog_info) {
 		@extra_args = ("--author=$maintainer", "--date=$date");
 	    }
+	    if ($signcommit) {
+		my $sign = '--gpg-sign';
+		$sign .= "=$keyid" if $keyid;
+		push(@extra_args, $sign);
+	    }
 	    $action_rc = action($prog, "commit", "-m", $message, @extra_args, @files_to_commit);
 	}
     }
@@ -581,11 +613,26 @@ sub commit {
 		my $time = Date::Format::strftime('%Y-%m-%d %H:%M:%S %z', \@time);
 		push(@extra_args, "--author=$maintainer", "--commit-time=$time");
 	    }
-	    $action_rc = action($prog, "commit", "-m", $message,
+	    my @sign_args;
+	    if ($signcommit) {
+		push(@sign_args, "-Ocreate_signatures=always");
+		if ($keyid) {
+		    push(@sign_args, "-Ogpg_signing_key=$keyid");
+		}
+	    }
+	    $action_rc = action($prog, @sign_args, "commit", "-m", $message,
 		@extra_args, @files_to_commit);
 	}
     }
     elsif ($prog eq 'darcs') {
+	if (! @files_to_commit && ($all || $release)) {
+	    # check to see if the WC is clean. darcs record would exit
+	    # nonzero, so don't run it in --all or --release mode.
+	    $action_rc = action($prog, "status");
+	    if (!$action_rc) {
+		    return;
+	    }
+	}
 	if ($diffmode) {
 	    $action_rc = action($prog, "diff", @files_to_commit);
 	} else {
@@ -604,18 +651,17 @@ sub commit {
 sub tag {
     my ($package, $tag) = @_;
 
+    # Make the message here so we can mangle $tag later, if needed
+    my $tag_msg = "tagging package $package version $tag";
     if ($prog eq 'svn' || $prog eq 'svk') {
 	my $svnpath=`svnpath`;
 	chomp $svnpath;
 	my $tagpath=`svnpath tags`;
 	chomp $tagpath;
 
-	if (! action($prog, "copy", $svnpath, "$tagpath/$tag",
-		     "-m", "tagging package $package version $tag")) {
-	    if (! action($prog, "mkdir", $tagpath,
-			 "-m", "create tag directory") ||
-		! action($prog, "copy", $svnpath, "$tagpath/$tag",
-			 "-m", "tagging package $package version $tag")) {
+	if (! action($prog, "copy", $svnpath, "$tagpath/$tag", "-m", $tag_msg)) {
+	    if (! action($prog, "mkdir", $tagpath, "-m", "create tag directory")
+		|| ! action($prog, "copy", $svnpath, "$tagpath/$tag", "-m", $tag_msg)) {
 		die "debcommit: failed tagging with $tag\n";
 	    }
 	}
@@ -649,47 +695,46 @@ sub tag {
 	    if (! action($prog, "tag", $tag)) {
 		die "debcommit: failed tagging with $tag\n";
 	    }
-        } else {
-		die "debcommit: bazaar or branch version too old to support tags\n";
-        }
+	} else {
+	    die "debcommit: bazaar or branch version too old to support tags\n";
+	}
     }
     elsif ($prog eq 'git') {
-	$tag=~s/^[0-9]+://; # strip epoch
-	$tag=~tr/~/./; # mangle for git
-	if ($tag=~/-/) {
+	$tag =~ tr/~/_/; # mangle for git
+	$tag =~ tr/:/%/;
+	if ($tag =~ /-/) {
 	    # not a native package, so tag as a debian release
-	    $tag="debian/$tag";
+	    $tag = "debian/$tag";
 	}
 
 	if ($signtags) {
-		if (defined $keyid) {
-			if (! action($prog, "tag", "-u", $keyid, "-m",
-			             "tagging package $package version $tag", $tag)) {
-	        		die "debcommit: failed tagging with $tag\n";
-			}
+	    my $tag_msg = "tagging package $package version $tag";
+	    if (defined $keyid) {
+		if (! action($prog, "tag", "-u", $keyid, "-m", $tag_msg, $tag)) {
+		    die "debcommit: failed tagging with $tag\n";
 		}
-		else {
-			if (! action($prog, "tag", "-s", "-m",
-			             "tagging package $package version $tag", $tag)) {
-	        		die "debcommit: failed tagging with $tag\n";
-			}
+	    }
+	    else {
+		if (! action($prog, "tag", "-s", "-m", $tag_msg, $tag)) {
+		    die "debcommit: failed tagging with $tag\n";
 		}
+	    }
 	}
 	elsif (! action($prog, "tag", $tag)) {
-	        die "debcommit: failed tagging with $tag\n";
-    	}
+	    die "debcommit: failed tagging with $tag\n";
+	}
     }
     elsif ($prog eq 'hg') {
-	$tag=~s/^[0-9]+://; # strip epoch
+	$tag =~ s/^[0-9]+://; # strip epoch
 	$tag="debian-$tag";
-	if (! action($prog, "tag", "-m", "tagging package $package version $tag", $tag)) {
-	        die "debcommit: failed tagging with $tag\n";
-    	}
+	if (! action($prog, "tag", "-m", $tag_msg, $tag)) {
+	    die "debcommit: failed tagging with $tag\n";
+	}
     }
     elsif ($prog eq 'darcs') {
-       if (! action($prog, "tag", $tag)) {
-               die "debcommit: failed tagging with $tag\n";
-       }
+	if (! action($prog, "tag", $tag)) {
+	    die "debcommit: failed tagging with $tag\n";
+	}
     }
     else {
 	die "debcommit: unknown program $prog";
