@@ -1,9 +1,8 @@
 #!/bin/sh
 
-# This program is designed to GPG sign a .dsc and .changes file pair
-# in the form needed for a legal Debian upload.  It is based in part
-# on dpkg-buildpackage.  It takes one argument: the name of the
-# .changes file.
+# This program is designed to GPG sign .dsc, .buildinfo, or .changes
+# files (or any combination of these) in the form needed for a legal
+# Debian upload.  It is based in part on dpkg-buildpackage.
 
 # Debian GNU/Linux debsign.  Copyright (C) 1999 Julian Gilbey.
 # Modifications to work with GPG by Joseph Carter and Julian Gilbey
@@ -19,7 +18,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
+# along with this program. If not, see <https://www.gnu.org/licenses/>.
 
 # Abort if anything goes wrong
 set -e
@@ -27,6 +26,8 @@ set -e
 PRECIOUS_FILES=0
 PROGNAME=`basename $0`
 MODIFIED_CONF_MSG='Default settings modified by devscripts configuration files:'
+HAVE_SIGNED=""
+NUM_SIGNED=0
 
 # Temporary directories
 signingdir=""
@@ -58,13 +59,13 @@ mkremotefilesdir () {
 
 usage () {
     echo \
-"Usage: debsign [options] [changes, dsc or commands file]
+"Usage: debsign [options] [changes, buildinfo, dsc or commands file]
   Options:
     -r [username@]remotehost
-                    The machine on which the changes/dsc files live.
-                    A changes file with full pathname (or relative
-                    to the remote home directory) must be given in
-                    such a case
+                    The machine on which the files live. If given, then a
+                    changes file with full pathname (or relative to the
+                    remote home directory) must be given as the main
+                    argument in the rest of the command line.
     -k<keyid>       The key to use for signing
     -p<sign-command>  The command to use for signing
     -e<maintainer>  Sign using key of <maintainer> (takes precedence over -m)
@@ -76,16 +77,16 @@ usage () {
     --re-sign       Re-sign if the file is already signed.
     --no-re-sign    Don't re-sign if the file is already signed.
     --debs-dir <directory>
-                    The location of the .changes / .dsc files when called from
+                    The location of the files to be signed when called from
                     within a source tree (default "..")
     --no-conf, --noconf
                     Don't read devscripts config files;
                     must be the first option given
     --help          Show this message
     --version       Show version and copyright information
-  If a commands or dsc or changes file is specified, it and any .dsc files in
-  the changes file are signed, otherwise debian/changelog is parsed to find
-  the changes file.
+  If an explicit filename is specified, it along with any child .buildinfo and
+  .dsc files are signed. Otherwise, debian/changelog is parsed to find the
+  changes file.
 
 $MODIFIED_CONF_MSG"
 }
@@ -114,6 +115,10 @@ temp_filename() {
     echo "$filename"
 }
 
+to_bool() {
+    if "$@"; then echo true; else echo false; fi
+}
+
 movefile() {
     if [ -w "$(dirname "$2")" ]; then
 	mv -f -- "$1" "$2"
@@ -156,11 +161,14 @@ mustsetvar () {
 # of dpkg-buildpackage, because we do not know all of the necessary
 # information when this function is read first.
 signfile () {
+    local type="$1"
+    local file="$2"
+    local signas="$3"
     local savestty=$(stty -g 2>/dev/null) || true
     mksigningdir
-    UNSIGNED_FILE="$signingdir/$(basename "$1")"
+    UNSIGNED_FILE="$signingdir/$(basename "$file")"
     ASCII_SIGNED_FILE="${UNSIGNED_FILE}.asc"
-    (cat "$1" ; echo "") > "$UNSIGNED_FILE"
+    (cat "$file" ; echo "") > "$UNSIGNED_FILE"
 
     gpgversion=`$signcommand --version | head -n 1 | cut -d' ' -f3`
     gpgmajorversion=`echo $gpgversion | cut -d. -f1`
@@ -168,7 +176,7 @@ signfile () {
 
     if [ $gpgmajorversion -gt 1 -o $gpgminorversion -ge 4 ]
     then
-	    $signcommand --local-user "$2" --clearsign \
+	    $signcommand --local-user "$signas" --clearsign \
 		--list-options no-show-policy-urls \
 		--armor --textmode --output "$ASCII_SIGNED_FILE"\
 		"$UNSIGNED_FILE" || \
@@ -178,7 +186,7 @@ signfile () {
 	      exit $SAVESTAT
 	    }
     else
-	    $signcommand --local-user "$2" --clearsign \
+	    $signcommand --local-user "$signas" --clearsign \
 		--no-show-policy-url \
 		--armor --textmode --output "$ASCII_SIGNED_FILE" \
 		"$UNSIGNED_FILE" || \
@@ -191,7 +199,9 @@ signfile () {
     stty $savestty 2>/dev/null || true
     echo
     PRECIOUS_FILES=$(($PRECIOUS_FILES + 1))
-    movefile "$ASCII_SIGNED_FILE" "$1"
+    HAVE_SIGNED="${HAVE_SIGNED:+${HAVE_SIGNED}, }$type"
+    NUM_SIGNED=$((NUM_SIGNED + 1))
+    movefile "$ASCII_SIGNED_FILE" "$file"
 }
 
 withecho () {
@@ -199,14 +209,24 @@ withecho () {
     "$@"
 }
 
+file_is_already_signed() {
+    test "$(head -n 1 "$1")" = "-----BEGIN PGP SIGNED MESSAGE-----"
+}
+
+unsignfile() {
+    UNSIGNED_FILE="$(temp_filename "$1" "unsigned")"
+
+    sed -e '1,/^$/d; /^$/,$d' "$1" > "$UNSIGNED_FILE"
+    movefile "$UNSIGNED_FILE" "$1"
+}
+
 # Has the dsc file already been signed, perhaps from a previous, partially
 # successful invocation of debsign?  We give the user the option of
 # resigning the file or accepting it as is.  Returns success if already
 # and failure if the file needs signing.  Parameters: $1=filename,
-# $2=file description for message (dsc or changes)
+# $2=file type for message (e.g. "changes", "commands")
 check_already_signed () {
-    [ "`head -n 1 \"$1\"`" = "-----BEGIN PGP SIGNED MESSAGE-----" ] || \
-	return 1
+    file_is_already_signed "$1" || return 1
 
     local resign
     if [ "$opt_re_sign" = "true" ]; then
@@ -228,10 +248,7 @@ check_already_signed () {
     [ "$resign" = "true" ] || \
 	return 0
 
-    UNSIGNED_FILE="$(temp_filename "$1" "unsigned")"
-
-    sed -e '1,/^$/d; /^$/,$d' "$1" > "$UNSIGNED_FILE"
-    movefile "$UNSIGNED_FILE" "$1"
+    withecho unsignfile "$1"
     return 1
 }
 
@@ -277,9 +294,6 @@ else
     # We do not replace this with a default directory to avoid accidentally
     # signing a broken package
     DEBRELEASE_DEBS_DIR="$(echo "${DEBRELEASE_DEBS_DIR%/}" | sed -e 's%/\+%/%g')"
-    if ! [ -d "$DEBRELEASE_DEBS_DIR" ]; then
-	debsdir_warning="config file specified DEBRELEASE_DEBS_DIR directory $DEBRELEASE_DEBS_DIR does not exist!"
-    fi
 
     # set config message
     MODIFIED_CONF=''
@@ -300,6 +314,7 @@ fi
 maint="$DEBSIGN_MAINT"
 signkey="$DEBSIGN_KEYID"
 debsdir="$DEBRELEASE_DEBS_DIR"
+debsdir_warning="config file specified DEBRELEASE_DEBS_DIR directory $DEBRELEASE_DEBS_DIR does not exist!"
 
 signcommand=''
 if [ -n "$DEBSIGN_PROGRAM" ]; then
@@ -307,6 +322,8 @@ if [ -n "$DEBSIGN_PROGRAM" ]; then
 else
     if command -v gpg > /dev/null 2>&1; then
 	signcommand=gpg
+    elif command -v gpg2 > /dev/null 2>&1; then
+	signcommand=gpg2
     fi
 fi
 
@@ -342,6 +359,7 @@ do
 	--debs-dir)
 	    shift
 	    opt_debsdir="$(echo "${1%/}" | sed -e 's%/\+%/%g')"
+	    debsdir_warning="could not find directory $opt_debsdir!"
 	    ;;
 	--no-conf|--noconf)
 		echo "$PROGNAME: $1 is only acceptable as the first command-line option!" >&2
@@ -356,22 +374,235 @@ do
 done
 
 debsdir=${opt_debsdir:-$debsdir}
-# check sanity of debsdir
-if ! [ -d "$debsdir" ]; then
-    if [ -n "$debsdir_warning" ]; then
-        echo "$PROGNAME: $debsdir_warning" >&2
-        exit 1
-    else
-        echo "$PROGNAME: could not find directory $debsdir!" >&2
-        exit 1
-    fi
-fi
 
 if [ -z "$signcommand" ]; then
     echo "Could not find a signing program!" >&2
     exit 1
 fi
 
+ensure_local_copy() {
+    local remotehost="$1"
+    local remotefile="$2"
+    local file="$3"
+    local type="$4"
+    if [ -n "$remotehost" ]
+    then
+	if [ ! -f "$file" ]
+	then
+	    withecho scp "$remotehost:$remotefile" "$file"
+	fi
+    fi
+
+    if [ ! -f "$file" -o ! -r "$file" ]
+    then
+	echo "$PROGNAME: Can't find or can't read $type file $file!" >&2
+	exit 1
+    fi
+}
+
+fixup_control() {
+    local filter_out="$1"
+    local childtype="$2"
+    local parenttype="$3"
+    local child="$4"
+    local parent="$5"
+    test -r "$child" || {
+	echo "$PROGNAME: Can't read .$childtype file $child!" >&2
+	return 1
+    }
+
+    local md5=$(md5sum "$child" | cut -d' ' -f1)
+    local sha1=$(sha1sum "$child" | cut -d' ' -f1)
+    local sha256=$(sha256sum "$child" | cut -d' ' -f1)
+    perl -i -pe 'BEGIN {
+    '" \$file=\"$child\"; \$md5=\"$md5\"; "'
+    '" \$sha1=\"$sha1\"; \$sha256=\"$sha256\"; "'
+    $size=(-s $file); ($base=$file) =~ s|.*/||;
+    $infiles=0; $inmd5=0; $insha1=0; $insha256=0; $format="";
+    }
+    if(/^Format:\s+(.*)/) {
+	$format=$1;
+	die "Unrecognised .$parenttype format: $format\n"
+	    unless $format =~ /^\d+(\.\d+)*$/;
+	($major, $minor) = split(/\./, $format);
+	$major+=0;$minor+=0;
+	die "Unsupported .$parenttype format: $format\n"
+	    if('"$filter_out"');
+    }
+    /^Files:/i && ($infiles=1,$inmd5=0,$insha1=0,$insha256=0);
+    if(/^Checksums-Sha1:/i) {$insha1=1;$infiles=0;$inmd5=0;$insha256=0;}
+    elsif(/^Checksums-Sha256:/i) {
+	$insha256=1;$infiles=0;$inmd5=0;$insha1=0;
+    } elsif(/^Checksums-Md5:/i) {
+	$inmd5=1;$infiles=0;$insha1=0;$insha256=0;
+    } elsif(/^Checksums-.*?:/i) {
+	die "Unknown checksum format: $_\n";
+    }
+    /^\s*$/ && ($infiles=0,$inmd5=0,$insha1=0,$insha256=0);
+    if ($infiles &&
+	/^ (\S+) (\d+) (\S+) (\S+) \Q$base\E\s*$/) {
+	$_ = " $md5 $size $3 $4 $base\n";
+	$infiles=0;
+    }
+    if ($inmd5 &&
+	/^ (\S+) (\d+) \Q$base\E\s*$/) {
+        $_ = " $md5 $size $base\n";
+        $inmd5=0;
+    }
+    if ($insha1 &&
+	/^ (\S+) (\d+) \Q$base\E\s*$/) {
+	$_ = " $sha1 $size $base\n";
+	$insha1=0;
+    }
+    if ($insha256 &&
+	/^ (\S+) (\d+) \Q$base\E\s*$/) {
+	$_ = " $sha256 $size $base\n";
+	$insha256=0;
+    }' "$parent"
+}
+
+fixup_buildinfo() {
+    fixup_control '($major != 0 or $minor > 2) and ($major != 1 or $minor > 0)' dsc buildinfo "$@"
+}
+
+fixup_changes() {
+    local childtype="$1"
+    shift
+    fixup_control '$major!=1 or $minor > 8 or $minor < 7' $childtype changes "$@"
+}
+
+withtempfile() {
+    local filetype="$1"
+    local mainfile="$2"
+    shift 2
+    local temp_file="$(temp_filename "$mainfile" "temp")"
+    cp "$mainfile" "$temp_file"
+    if "$@" "$temp_file"; then
+	if ! cmp -s "$mainfile" "$temp_file"; then
+	    # emulate output of "withecho" but on the mainfile
+	    echo " $@" "$mainfile" >&2
+	fi
+	movefile "$temp_file" "$mainfile"
+    else
+	rm "$temp_file"
+	echo "$PROGNAME: Error processing .$filetype file (see above)" >&2
+	exit 1
+    fi
+}
+
+guess_signas() {
+    if [ -n "$maint" ]
+    then maintainer="$maint"
+    # Try the new "Changed-By:" field first
+    else maintainer=`sed -n 's/^Changed-By: //p' $1`
+    fi
+    if [ -z "$maint" ]
+    then maintainer=`sed -n 's/^Maintainer: //p' $1`
+    fi
+
+    echo "${signkey:-$maintainer}"
+}
+
+maybesign_dsc() {
+    local signas="$1"
+    local remotehost="$2"
+    local remotedsc="$3"
+    local dsc="$4"
+
+    ensure_local_copy "$remotehost" "$remotedsc" "$dsc" dsc
+    if check_already_signed "$dsc" dsc; then
+	echo "Leaving current signature unchanged." >&2
+	return
+    fi
+
+    withecho signfile dsc "$dsc" "$signas"
+
+    if [ -n "$remotehost" ]
+    then
+	withecho scp "$dsc" "$remotehost:$remotedsc"
+	PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
+    fi
+}
+
+maybesign_buildinfo() {
+    local signas="$1"
+    local remotehost="$2"
+    local remotebuildinfo="$3"
+    local buildinfo="$4"
+    local remotedsc="$5"
+    local dsc="$6"
+
+    ensure_local_copy "$remotehost" "$remotebuildinfo" "$buildinfo" buildinfo
+    if check_already_signed "$buildinfo" "buildinfo"; then
+       echo "Leaving current signature unchanged." >&2
+       return
+    fi
+
+    if grep -q `basename "$dsc"` "$buildinfo"; then
+	maybesign_dsc "$signas" "$remotehost" "$remotedsc" "$dsc"
+	withtempfile buildinfo "$buildinfo" fixup_buildinfo "$dsc"
+    fi
+
+    withecho signfile buildinfo "$buildinfo" "$signas"
+
+    if [ -n "$remotehost" ]
+    then
+	withecho scp "$buildinfo" "$remotehost:$remotebuildinfo"
+	PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
+    fi
+}
+
+maybesign_changes() {
+    local signas="$1"
+    local remotehost="$2"
+    local remotechanges="$3"
+    local changes="$4"
+    local remotebuildinfo="$5"
+    local buildinfo="$6"
+    local remotedsc="$7"
+    local dsc="$8"
+
+    ensure_local_copy "$remotehost" "$remotechanges" "$changes" changes
+    if check_already_signed "$changes" "changes"; then
+	echo "Leaving current signature unchanged." >&2
+	return
+    fi
+
+    hasdsc="$(to_bool grep -q `basename "$dsc"` "$changes")"
+    hasbuildinfo="$(to_bool grep -q `basename "$buildinfo"` "$changes")"
+
+    if $hasbuildinfo; then
+	# assume that this will also sign the same dsc if it's available
+	maybesign_buildinfo "$signas" "$remotehost" \
+	    "$remotebuildinfo" "$buildinfo" \
+	    "$remotedsc" "$dsc"
+    elif $hasdsc; then
+	maybesign_dsc "$signas" "$remotehost" "$remotedsc" "$dsc"
+    fi
+
+    if $hasdsc; then
+	withtempfile changes "$changes" fixup_changes dsc "$dsc"
+    fi
+    if $hasbuildinfo; then
+	withtempfile changes "$changes" fixup_changes buildinfo "$buildinfo"
+    fi
+    withecho signfile changes "$changes" "$signas"
+
+    if [ -n "$remotehost" ]
+    then
+	withecho scp "$changes" "$remotehost:$remotechanges"
+	PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
+    fi
+}
+
+report_signed() {
+    if [ $NUM_SIGNED -eq 1 ]; then
+	echo "Successfully signed $HAVE_SIGNED file"
+    elif [ $NUM_SIGNED -gt 0 ]; then
+	echo "Successfully signed $HAVE_SIGNED files"
+    fi
+}
+
 dosigning() {
     # Do we have to download the changes file?
     if [ -n "$remotehost" ]
@@ -380,29 +611,22 @@ dosigning() {
 	cd "$remotefilesdir"
 
 	remotechanges=$changes
+	remotebuildinfo=$buildinfo
 	remotedsc=$dsc
 	remotecommands=$commands
 	remotedir="`perl -e 'chomp($_="'"$dsc"'"); m%/% && s%/[^/]*$%% && print'`"
 	changes=`basename "$changes"`
+	buildinfo=`basename "$buildinfo"`
 	dsc=`basename "$dsc"`
 	commands=`basename "$commands"`
 
-	if [ -n "$changes" ]
-	then
-	    if [ ! -f "$changes" ]
-	    then
-		withecho scp "$remotehost:$remotechanges" .
-	    fi
-	elif [ -n "$dsc" ]
-	then withecho scp "$remotehost:$remotedsc" "$dsc"
-	else withecho scp "$remotehost:$remotecommands" "$commands"
-	fi
-
 	if [ -n "$changes" ] && echo "$changes" | egrep -q '[][*?]'
 	then
+	    withecho scp "$remotehost:$remotechanges" .
 	    for changes in $changes
 	    do
 		printf "\n"
+		buildinfo="${remotedir+$remotedir/}${changes%.changes}.buildinfo"
 		dsc=`echo "${remotedir+$remotedir/}$changes" | \
 		    perl -pe 's/\.changes$/.dsc/; s/(.*)_(.*)_(.*)\.dsc/\1_\2.dsc/'`
 		dosigning;
@@ -411,116 +635,7 @@ dosigning() {
 	fi
     fi
 
-    if [ -n "$changes" ]
-    then
-	if [ ! -f "$changes" -o ! -r "$changes" ]
-	then
-	    echo "$PROGNAME: Can't find or can't read changes file $changes!" >&2
-	    exit 1
-	fi
-
-	check_already_signed "$changes" "changes" && {
-	   echo "Leaving current signature unchanged." >&2
-	   return
-	}
-	if [ -n "$maint" ]
-	then maintainer="$maint"
-	# Try the "Changed-By:" field first
-	else maintainer=`sed -n 's/^Changed-By: //p' $changes`
-	fi
-	if [ -z "$maintainer" ]
-	then maintainer=`sed -n 's/^Maintainer: //p' $changes`
-	fi
-
-	signas="${signkey:-$maintainer}"
-
-	# Is there a dsc file listed in the changes file?
-	if grep -q `basename "$dsc"` "$changes"
-	then
-	    if [ -n "$remotehost" ]
-	    then
-		withecho scp "$remotehost:$remotedsc" "$dsc"
-	    fi
-
-	    if [ ! -f "$dsc" -o ! -r "$dsc" ]
-	    then
-		echo "$PROGNAME: Can't find or can't read dsc file $dsc!" >&2
-		exit 1
-	    fi
-	    check_already_signed "$dsc" "dsc" || withecho signfile "$dsc" "$signas"
-	    dsc_md5=`md5sum $dsc | cut -d' ' -f1`
-	    dsc_sha1=`sha1sum $dsc | cut -d' ' -f1`
-	    dsc_sha256=`sha256sum $dsc | cut -d' ' -f1`
-
-	    temp_changes="$(temp_filename "$changes" "temp")"
-	    cp "$changes" "$temp_changes"
-	    if perl -i -pe 'BEGIN {
-		'" \$dsc_file=\"$dsc\"; \$dsc_md5=\"$dsc_md5\"; "'
-		'" \$dsc_sha1=\"$dsc_sha1\"; \$dsc_sha256=\"$dsc_sha256\"; "'
-		$dsc_size=(-s $dsc_file); ($dsc_base=$dsc_file) =~ s|.*/||;
-		$infiles=0; $insha1=0; $insha256=0; $format="";
-		}
-		if(/^Format:\s+(.*)/) {
-		    $format=$1;
-		    die "Unrecognised .changes format: $format\n"
-			unless $format =~ /^\d+(\.\d+)*$/;
-		    ($major, $minor) = split(/\./, $format);
-		    $major+=0;$minor+=0;
-		    die "Unsupported .changes format: $format\n"
-			if($major!=1 or $minor > 8 or $minor < 7);
-		}
-		/^Files:/i && ($infiles=1,$insha1=0,$insha256=0);
-		if(/^Checksums-Sha1:/i) {$insha1=1;$infiles=0;$insha256=0;}
-		elsif(/^Checksums-Sha256:/i) {
-		    $insha256=1;$infiles=0;$insha1=0;
-		} elsif(/^Checksums-.*?:/i) {
-		    die "Unknown checksum format: $_\n";
-		}
-		/^\s*$/ && ($infiles=0,$insha1=0,$insha256=0);
-		if ($infiles &&
-		    /^ (\S+) (\d+) (\S+) (\S+) \Q$dsc_base\E\s*$/) {
-		    $_ = " $dsc_md5 $dsc_size $3 $4 $dsc_base\n";
-		    $infiles=0;
-		}
-		if ($insha1 &&
-		    /^ (\S+) (\d+) \Q$dsc_base\E\s*$/) {
-		    $_ = " $dsc_sha1 $dsc_size $dsc_base\n";
-		    $insha1=0;
-		}
-		if ($insha256 &&
-		    /^ (\S+) (\d+) \Q$dsc_base\E\s*$/) {
-		    $_ = " $dsc_sha256 $dsc_size $dsc_base\n";
-		    $insha256=0;
-		}' "$temp_changes"
-	    then
-		movefile "$temp_changes" "$changes"
-	    else
-		rm "$temp_changes"
-		echo "$PROGNAME: Error processing .changes file (see above)" >&2
-		exit 1
-	    fi
-
-	    withecho signfile "$changes" "$signas"
-
-	    if [ -n "$remotehost" ]
-	    then
-		withecho scp "$changes" "$dsc" "$remotehost:$remotedir"
-		PRECIOUS_FILES=$(($PRECIOUS_FILES - 2))
-	    fi
-
-	    echo "Successfully signed dsc and changes files"
-	else
-	    withecho signfile "$changes" "$signas"
-
-	    if [ -n "$remotehost" ]
-	    then
-		withecho scp "$changes" "$remotehost:$remotedir"
-		PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
-	    fi
-
-	    echo "Successfully signed changes file"
-	fi
-    elif [ -n "$commands" ] # sign .commands file
+    if [ -n "$commands" ] # sign .commands file
     then
 	if [ ! -f "$commands" -o ! -r "$commands" ]
 	then
@@ -528,6 +643,7 @@ dosigning() {
 	    exit 1
 	fi
 
+	ensure_local_copy "$remotehost" "$remotecommands" "$commands" commands
 	check_already_signed "$commands" commands && {
 	    echo "Leaving current signature unchanged." >&2
 	    return
@@ -579,7 +695,7 @@ for valid format" >&2;
 
 	signas="${signkey:-$maintainer}"
 
-	withecho signfile "$commands" "$signas"
+	withecho signfile commands "$commands" "$signas"
 
 	if [ -n "$remotehost" ]
 	then
@@ -587,38 +703,31 @@ for valid format" >&2;
 	    PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
 	fi
 
-	echo "Successfully signed commands file"
-    else # only a dsc file to sign; much easier
-	if [ ! -f "$dsc" -o ! -r "$dsc" ]
-	then
-	    echo "$PROGNAME: Can't find or can't read dsc file $dsc!" >&2
-	    exit 1
-	fi
+	report_signed
 
-	check_already_signed "$dsc" dsc && {
-	    echo "Leaving current signature unchanged." >&2
-	    return
-	}
-	if [ -n "$maint" ]
-	then maintainer="$maint"
-	# Try the new "Changed-By:" field first
-	else maintainer=`sed -n 's/^Changed-By: //p' $dsc`
-	fi
-	if [ -z "$maint" ]
-	then maintainer=`sed -n 's/^Maintainer: //p' $dsc`
-	 fi
+    elif [ -n "$changes" ]
+    then
+	signas="$(guess_signas "$changes")"
+	maybesign_changes "$signas" "$remotehost" \
+	    "$remotechanges" "$changes" \
+	    "$remotebuildinfo" "$buildinfo" \
+	    "$remotedsc" "$dsc"
+	report_signed
 
-	signas="${signkey:-$maintainer}"
+    elif [ -n "$buildinfo" ]
+    then
+	signas="$(guess_signas "$buildinfo")"
+	maybesign_buildinfo "$signas" "$remotehost" \
+	    "$remotebuildinfo" "$buildinfo" \
+	    "$remotedsc" "$dsc"
+	report_signed
 
-	withecho signfile "$dsc" "$signas"
+    else
+	signas="$(guess_signas "$dsc")"
+	maybesign_dsc "$signas" "$remotehost" \
+	    "$remotedsc" "$dsc"
+	report_signed
 
-	if [ -n "$remotehost" ]
-	then
-	    withecho scp "$dsc" "$remotehost:$remotedsc"
-	    PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
-	fi
-
-	echo "Successfully signed dsc file"
     fi
 }
 
@@ -627,8 +736,13 @@ for valid format" >&2;
 # out its name from debian/changelog
 case $# in
     0)	# We have to parse debian/changelog to find the current version
+	# check sanity of debsdir
+	if ! [ -d "$debsdir" ]; then
+	    echo "$PROGNAME: $debsdir_warning" >&2
+	    exit 1
+	fi
 	if [ -n "$remotehost" ]; then
-	    echo "$PROGNAME: Need to specify a .changes, .dsc or .commands file location with -r!" >&2
+	    echo "$PROGNAME: Need to specify a remote file location when giving -r!" >&2
 	    exit 1
 	fi
 	if [ ! -r debian/changelog ]; then
@@ -636,14 +750,20 @@ case $# in
 	    exit 1
 	fi
 
-	mustsetvar package "`dpkg-parsechangelog | sed -n 's/^Source: //p'`" \
-	    "source package"
-	mustsetvar version "`dpkg-parsechangelog | sed -n 's/^Version: //p'`" \
-	    "source version"
+	mustsetvar package "`dpkg-parsechangelog -SSource`" "source package"
+	mustsetvar version "`dpkg-parsechangelog -SVersion`" "source version"
 
 	if [ "x$sourceonly" = x ]
 	then
-	    mustsetvar arch "`dpkg-architecture -a${targetarch} -t${targetgnusystem} -qDEB_HOST_ARCH`" "build architecture"
+	    if [ -n "$targetarch" ] && [ -n "$targetgnusystem" ]; then
+		mustsetvar arch "$(dpkg-architecture "-a${targetarch}" "-t${targetgnusystem}" -qDEB_HOST_ARCH)" "build architecture"
+	    elif [ -n "$targetarch" ]; then
+		mustsetvar arch "$(dpkg-architecture "-a${targetarch}" -qDEB_HOST_ARCH)" "build architecture"
+	    elif [ -n "$targetgnusystem" ]; then
+		mustsetvar arch "$(dpkg-architecture "-t${targetgnusystem}" -qDEB_HOST_ARCH)" "build architecture"
+	    else
+		mustsetvar arch "$(dpkg-architecture -qDEB_HOST_ARCH)" "build architecture"
+	    fi
 	else
 	    arch=source
 	fi
@@ -652,9 +772,11 @@ case $# in
 	pv="${package}_${sversion}"
 	pva="${package}_${sversion}_${arch}"
 	dsc="$debsdir/$pv.dsc"
+	buildinfo="$debsdir/$pva.buildinfo"
 	changes="$debsdir/$pva.changes"
 	if [ -n "$multiarch" -o ! -r $changes ]; then
 	    changes=$(ls "$debsdir/${package}_${sversion}_*+*.changes" "$debsdir/${package}_${sversion}_multi.changes" 2>/dev/null | head -1)
+	    # TODO: dpkg-cross does not yet do buildinfo, so don't worry about it here
 	    if [ -z "$multiarch" ]; then
 		if [ -n "$changes" ]; then
 		    echo "$PROGNAME: could not find normal .changes file but found multiarch file:" >&2
@@ -677,22 +799,32 @@ case $# in
 	    case "$1" in
 		*.dsc)
 		    changes=
+		    buildinfo=
 		    dsc=$1
 		    commands=
 		    ;;
+	        *.buildinfo)
+		    changes=
+		    buildinfo=$1
+		    dsc=`echo $buildinfo | \
+			perl -pe 's/\.buildinfo$/.dsc/; s/(.*)_(.*)_(.*)\.dsc/\1_\2.dsc/'`
+		    commands=
+		    ;;
 	        *.changes)
 		    changes=$1
+		    buildinfo="${changes%.changes}.buildinfo"
 		    dsc=`echo $changes | \
 			perl -pe 's/\.changes$/.dsc/; s/(.*)_(.*)_(.*)\.dsc/\1_\2.dsc/'`
 		    commands=
 		    ;;
 		*.commands)
 		    changes=
+		    buildinfo=
 		    dsc=
 		    commands=$1
 		    ;;
 		*)
-		    echo "$PROGNAME: Only a .changes, .dsc or .commands file is allowed as argument!" >&2
+		    echo "$PROGNAME: Only a .changes, .buildinfo, .dsc or .commands file is allowed as argument!" >&2
 		    exit 1 ;;
 	    esac
 	    dosigning
